import cognitojwt
import logging
from cognitojwt import jwt_sync
import boto3
import traceback
import json
from flask import request
from functools import wraps
from flask_api import status as http_status
import copy
#from cryptography.fernet import Fernet
from .connect_database import ConnectDatabase
from .response_json import Response

logging = logging.getLogger('app.sub')


class Library(object):
    def __init__(self, cognito_region='eu-west-1',
                 sql_settings=None,
                 standard_roles=None,
                 mongo_url=None,
                 mongo_db=None,
                 app_client_id=None):
        # Declaring global variables

        self.app_client_id = app_client_id                   # app_client_id of AWS cognito
        self.mongo_url = mongo_url                           # Mongo database url
        self.mongo_db = mongo_db                             # Mongo database name
        self.standard_roles = standard_roles                 # standard roles initialised
        self.cognito_region = cognito_region                 # Aws cognito region
        self.data_token = None                               # data_token it is request payload
        self.username = None                                 # cognito username
        self.sql_settings = sql_settings                     # sql database configurations
        self.role = None                                     # user role in organisation
        self.role_permissions = {}                           # User role permissions set
        self.message_access_denied = "access_denied"         # return message for access denied
        self.process_failed_message = "failed_to_process"    # return message for process failure
        self.user_access_permissions = None                 # user permissions for respected tasks
        self.request_service_payment = None                 # To upgrade subscription
        self.response_obj = Response()                       # template for return response
        self.connect_db_obj = ConnectDatabase(
            self.sql_settings, self.mongo_url)               # creating object for database connection

    def get_ssm_param(self, paramname, ssm_client="ssm"):
        """
        Info: It hits aws ssm parameter store and returns the requested parameter value.
        param paramname: paramname should be the ssm parameter name.
        return: It returns ssm value which is assigned to ssm parameter.

        """
        try:
            session = boto3.Session(region_name=self.cognito_region)
            ssm = session.client(ssm_client)
            paramdata = ssm.get_parameter(Name=paramname, WithDecryption=True)
            return self.response_obj.output_response(data=format(paramdata['Parameter']['Value']))
        except Exception as error:
            return self.response_obj.output_response(error=False, status=500, data=str(error))

    def user_authorisation(self, userpool_id, testmode, auth_type="full"):
        """
        Info: It validate cognito token valididty and user existence
        :param userpool_id: cognito userpool id
        :param testmode: Boolean, If it is True then validation
        :param auth_type: Type of authentication full/aws/account
        ignores token expiration else check for token expiration
        :return:
        """
        def authorisation_check(f):
            """
            Info: Decorator which authorise aws token
            :param f: Function
            :return:
            """
            @wraps(f)
            def decorated_function(*args, **kws):

                # Extracting header from the request

                logging.info("getting token from Header")
                token = request.headers['Authorization']
                logging.info(token)

                # Calling aws_authorisation function for validating aws token
                if auth_type != "account":

                    logging.info("processing aws authorisation")
                    aws_authorise = Library.aws_authorisation(
                        self, token, userpool_id, testmode)
                    logging.info("aws authorisation completed")

                    # Making sure that the function not returned any error
                    # aborting the process if any error occurs

                    if aws_authorise[self.response_obj.k_status] != 200:
                        return aws_authorise, aws_authorise[self.response_obj.k_status]

                # calling valid_user_check function to validate given aws
                # token and account_id referring same person or not
                if auth_type != "aws":

                    logging.info("user validation started by using"
                                 " account_id and cognito username")

                    valid_user = Library.valid_user_check(self)
                    logging.info("user validation completed")

                    # If user not valid then aborting the process by giving return message

                    if valid_user[self.response_obj.k_data] is not True:
                        logging.info(valid_user)
                        logging.info("invalid user, aborting the process")
                        return valid_user, valid_user[self.response_obj.k_status]

                    logging.info("found user is valid processing further")

                return f(*args, **kws)

            return decorated_function

        return authorisation_check

    def aws_authorisation(self, token, userpool_id, testmode):
        """
        Info: Validate AWS token is valid or not
        :param token: Aws token
        :param userpool_id: AWS user pool id
        :param testmode: Boolean, If it is True then validation
        ignores token expiration else check for token expiration
        :return:
        """
        try:

            # Validating aws token by using jwt_sync library

            logging.info("aws authorisation started")
            data = jwt_sync.decode(token, self.cognito_region,
                                   userpool_id, self.app_client_id,
                                   testmode=testmode)

            # Getting username from the data after decoding aws token

            self.username = data['username']

            # returning status 200

            return self.response_obj.output_response()

        # Catching aws token expiry exception

        except cognitojwt.exceptions.CognitoJWTException as error:
            logging.info("exception occurred while processing "
                         "aws authorisation : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_401_UNAUTHORIZED,
                message="authorization_header_token_expired",
                data=str(error))

        # Catching any exception occurs in the process

        except Exception as error:
            logging.info("exception occurred while processing "
                         "aws authorisation : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message="authorization_header_token_is_invalid",
                data=str(error))

    def valid_user_check(self):
        """
        Info: Validate whether cognito username is belong to the account_id provided in payload
        :return: returns data True if authentication success
        """
        try:
            # Extracting payload from the request

            self.data_token = json.loads(request.form["payload"])
            logging.info(self.data_token)

            # Authenticating user by using cognito username and account_id

            result = self.connect_db_obj.execute_procedure(
                'user_authentication',
                (self.username, self.data_token['account_id']),
                "select",
                True)

            # Making sure that execution of stored procedure not returned any error

            if result[self.response_obj.k_status] != 200:
                logging.info(result[self.response_obj.k_data])
                return result

            # Fetching result from procedure execution

            valid_user = result[self.response_obj.k_data][0][0]

            # procedure returns 1 if user is valid

            if valid_user == 1:
                # returning True if user is valid
                return self.response_obj.output_response(data=True)

            # User is invalid means cognito username and account_id are not related to one person

            else:
                logging.info("user is invalid, So aborting process")
                return self.response_obj.output_response(
                            error=True,
                            status=http_status.HTTP_403_FORBIDDEN,
                            message=self.message_access_denied)

        # Catching any exception occurs in the process

        except Exception as error:
            logging.info("exception occurred while validating user : {}"
                         "".format(str(traceback.format_exc())))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message="authorization_header_token_is_invalid",
                data=str(error))

    def subscription_check(self, mongo_collection, subscription_dict):

        def check_subscription(f):
            @wraps(f)
            def decorated_function(*args, **kws):
                try:
                    # Getting subscription template

                    logging.info("getting subscription document from mongo")
                    collection = self.connect_db_obj.connect_mongo_db(
                        self.mongo_db, mongo_collection)
                    if collection[self.response_obj.k_status] != 200:
                        return collection
                    logging.info(collection)
                    logging.info(self.data_token["subscription_type"])
                    data = collection[self.response_obj.k_data].find_one(
                        {"subscription_type": self.data_token["subscription_type"]})
                    logging.info("got subscriptions from the database")

                    for k, v in subscription_dict.items():
                        if data[k][v] == "true":
                            self.request_service_payment = False
                            return f(*args, **kws)
                        elif data[k][v] == "false":
                            status = http_status.HTTP_403_FORBIDDEN
                            message = "access_denied"
                        elif data[k][v] == "upgrade":
                            status = http_status.HTTP_409_CONFLICT
                            message = "upgrade_plan"
                        elif data[k][v] == "request_service_payment":
                            self.request_service_payment = True
                            return f(*args, **kws)
                        else:
                            status = http_status.HTTP_500_INTERNAL_SERVER_ERROR
                            message = "invalid_data"
                        break

                    # Returning response

                    return self.response_obj.output_response(
                        error=True,
                        status=status,
                        message=message), status

                # Catching any exception occurs in the process

                except Exception as error:
                    logging.info("exception occurred while checking "
                                 "subscription : {}".format(error))
                    return self.response_obj.output_response(
                        error=True,
                        status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                        message=self.process_failed_message,
                        data=str(error)), http_status.HTTP_500_INTERNAL_SERVER_ERROR

            return decorated_function

        return check_subscription

    def role_policy_check(self, mongo_org_db,
                          mongo_role_collection,
                          action: dict = None):
        logging.info(action)

        def permission_check(f):
            @wraps(f)
            def decorated_function(*args, **kws):
                try:
                    # Getting role permissions

                    self.user_access_permissions = None
                    logging.info("getting role permissions")
                    response = self.get_role_permissions(mongo_org_db, mongo_role_collection)

                    # Making sure that method not returns any error

                    if response[self.response_obj.k_status] != 200:
                        return response, response[self.response_obj.k_status]

                    # If action is None processing further

                    if action is None:
                        return f(*args, **kws)

                    #  Fetching user access permissions and view org permissions

                    if self.role_permissions != {} and action is not None:
                        logging.info("getting user access permissions")
                        response = self.policy_check(
                            action, self.role_permissions)
                        if response[self.response_obj.k_status] != 200:
                            return response
                        logging.info("got user access permissions")
                        self.user_access_permissions = response[self.response_obj.k_data]
                        logging.info(self.user_access_permissions)

                    # Processing further if user have any of the permission

                    if self.user_access_permissions is not None or self.role == 'Admin':
                        return f(*args, **kws)

                    # User not have enough permissions so aborting process
                    else:
                        return self.response_obj.output_response(
                            error=True,
                            status=http_status.HTTP_403_FORBIDDEN,
                            message=self.message_access_denied), http_status.HTTP_403_FORBIDDEN

                # Catching any exception occurs in the process

                except Exception as error:
                    logging.info("exception occurred while checking "
                                 "role policies : {}".format(error))
                    return self.response_obj.output_response(
                        error=True,
                        status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                        message=self.process_failed_message,
                        data=str(error)), http_status.HTTP_500_INTERNAL_SERVER_ERROR

            return decorated_function

        return permission_check

    def get_role_permissions(self, mongo_org_db, mongo_role_collection):
        """
        Info: Get role permissions
        :param mongo_org_db: Mongo database name where organisation data stored
        :param mongo_role_collection: Mongo collection name where organisation roles stored
        :return:
        """
        try:
            # Getting user role name based on user_id and org_id

            self.role_permissions = {}
            logging.info("getting role name")
            response = self.connect_db_obj.execute_procedure(
                'get_role_name',
                (self.data_token["user_id"],
                 self.data_token["org_id"]),
                "select",
                True)
            logging.info("got role name")

            # Making sure that procedure execution not resulted any error
            if response[self.response_obj.k_status] != 200:
                return response

            # If procedure returns 0 then there is a mismatch between user_id and org_id

            if len(response[self.response_obj.k_data]) == 0:
                logging.info("found there is a mismatch between user_id and org_id")
                return self.response_obj.output_response(
                    error=False,
                    status=http_status.HTTP_403_FORBIDDEN,
                    message=self.message_access_denied)

            # Fetching role name from procedure response

            self.role = response[self.response_obj.k_data][0][0]

            # User role is not in the standard roles list

            if self.role != 'Admin':

                # Fetching role permissions from mongo database

                logging.info("fetching role permissions from mongo")
                collection_roles = self.connect_db_obj.connect_mongo_db(
                    mongo_org_db, mongo_role_collection)
                if collection_roles[self.response_obj.k_status] != 200:
                    return collection_roles
                data = collection_roles[self.response_obj.k_data].find_one(
                    {"org_id": self.data_token['org_id'], "role_name": self.role})
                self.role_permissions = data["role_permissions"]
                logging.info("got role permissions")

            # Returning status 200
            return self.response_obj.output_response()

        # Catching any exception occurs in the process and aborting process

        except Exception as error:
            logging.info("exception occured while getting role "
                         "permissions : {}".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    def policy_check(self, action, role_permission_set):
        try:
            logging.info(action)
            policy_dict = {}
            temp_action = copy.deepcopy(action)

            # If Organisation key is exist in action then get values as permission set

            if "administration" in action.keys():
                i = "Organisation"

                # Getting common elements between given set of permissions and role permissions

                key = list(action[i].keys())[0]
                response = Library.common_dict_items(self, action[i][key], role_permission_set[i][key])
                logging.info(response)

                # Making sure that the function does not return any error
                if response[self.response_obj.k_status] != 200:
                    return response

                # Got permission set which is common between action and role permissions
                elif response[self.response_obj.k_data] != {}:
                    # Updating policy dict
                    policy_dict[i][key] = response[self.response_obj.k_data]

                # Deleting dictionary data based on organisation key
                del temp_action[i]

            # Checking whether action has any elements
            if len(temp_action) > 0:

                # Getting common elements between given set of permissions and role permissions

                key = list(temp_action.keys())[0]
                response = Library.common_dict_items(self, temp_action[key], role_permission_set[key])

                # Making sure that the function does not return any error
                if response[self.response_obj.k_status] != 200:
                    return response

                # Got permission set which is common between temp_action and role permissions
                elif response[self.response_obj.k_data] != {}:
                    # Updating policy dict
                    policy_dict[key]=response[self.response_obj.k_data]

            # Checking whether policy dict is empty then returning default response with status 200
            if policy_dict == {}:
                return self.response_obj.output_response()

            # Policy dict is not empty so including data in the response
            else:
                return self.response_obj.output_response(data=policy_dict)

        # Catching any exception if occurs in the process

        except Exception as error:
            logging.info("exception occured while running policy_check : {}"
                         "".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    def common_dict_items(self, dict1, dict2):
        """
        Info: Get common elements in two dictionaries
        :param dict1: Dictionary 1
        :param dict2: Dictionary 2
        :return: Common elements between dict1 and dict2
        """
        try:
            dict1.update((k, dict2[k]) for k in set(dict2).intersection(dict1))
            dict1 = {k: dict1[k] for k in dict1 if dict1[k] == "true"}
            return self.response_obj.output_response(data=dict1)
        except Exception as error:
            logging.info("exception occured while running common_dict_items : {}"
                         "".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    # static template for Organisation level view permissions
    @staticmethod
    def view_org_permission_set():
        permission_set = {
            "Organisation": {
                "Organisation Profile": ["View"],
                "Subscription": ["View"]
            },
            "Users": ["View"],
            "Roles": ["View"]
        }
        return permission_set

    def get_users(self):
        """
        Info: Fetch organisation level users from the database
        :return: Users data
        """
        try:
            # Executing stored procedure to get users from database
            response = self.connect_db_obj.execute_procedure(
                'get_users',
                (self.data_token["org_id"],),
                'select', True)

            # Check for any error response
            if response[self.response_obj.k_status] != 200:
                return self.response_obj.output_response(
                    error=False,
                    status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                    message=self.process_failed_message)

            # Preparing list of users from json object
            user_list = []
            for user in response[self.response_obj.k_data]:
                user = json.loads(user[0])
                if user["user_status"] == "Active":
                    user_list.append(user)

            # Returning users list
            return self.response_obj.output_response(data=user_list)

        # Catching any exception if occurs in the process
        except Exception as error:
            logging.info("exception occurred while fetching users : {}"
                         "".format(error))
            return self.response_obj.output_response(
                error=True,
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                message=self.process_failed_message,
                data=str(error))

    def get_mail(self, procedure, user_id):
        try:
            logging.info("Adding log in company log table")
            mail = self.connect_db_obj.execute_procedure(
                procedure,
                (user_id,), 'select',
                True)

            if mail[self.response_obj.k_status] != 200:
                return mail
            mail = mail[self.response_obj.k_data][0][0]
            return self.response_obj.output_response(data=mail)
        except Exception as error:
            logging.info(error)
            return self.response_obj.output_response(
                status=http_status.HTTP_500_INTERNAL_SERVER_ERROR,
                error=True,
                data=str(error))

    @staticmethod
    def aws_s3(key, s3_bucket):
        session = boto3.Session(region_name='eu-west-1')
        s3_session = session.resource('s3')

        # get a handle on the bucket that holds your file
        bucket = s3_session.Bucket(s3_bucket)  # example: energy_market_procesing

        # get a handle on the object you want (i.e. your file)
        obj = bucket.Object(key=key)  # example: market/zone1/data.csv

        # get the object
        response = obj.get()

        # read the contents of the file
        lines = response['Body'].read().decode(encoding="utf-8", errors="ignore")
        return lines

    @staticmethod
    def view_company_policy_check(f):
        @wraps(f)
        def decorated_function(*args, **kws):
            logging.info("check company policies")
            return f(*args, **kws)

        return decorated_function

    # @staticmethod
    # def encrypt(secrete_key, string):
    #     """
    #     Info: Encrypt given string using secrete key
    #     :param secrete_key: Secrete key for encoding and decoding script
    #     :param string: String to encode
    #     :return: Encoded string
    #     """
    #     key = secrete_key.encode(encoding='utf-8')
    #     cipher_suite = Fernet(key)
    #     encoded_text = cipher_suite.encrypt(string.encode(encoding='utf-8')).decode("utf-8")
    #     return encoded_text

    @staticmethod
    def company_profile_data():
        return {
            "org_id": "",
            "company_number": "",
            "company_profile_id": "",
            "company_profile_status": "",
            "assignment": [
                {
                    "profile_id": "",
                    "user_status": "",
                    "account_id": "",
                    "imported_user": ""
                }]}

    @staticmethod
    def company_data():
        return {"Jurisdiction": "",
                "TradingOnMarket": "",
                "DTR5Applies": "",
                "RegisteredOfficeAddress": "",
                "SailAddress": "",
                "IncorporationDate": "",
                "OriginatingRegistry": {
                    "country": ""
                },
                "Accounts": {
                    "next_made_up_to": "",
                    "next_due": ""
                },
                "ConfirmationStatement": {
                    "next_made_up_to": "",
                    "next_due": ""
                }
                }
